/* /////////////////////////////////////////////////////////////////////////////
 * File:        rdempty.cpp
 *
 * Purpose:     Implementation file for the rdempty project.
 *
 * Created:     21st August 2003
 * Updated:     21st May 2005
 *
 * Status:      Wizard-generated
 *
 * License:     (Licensed under the Synesis Software Open License)
 *
 *              Copyright 2003-2005, Synesis Software Pty Ltd.
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/software
 *
 *              email:      software@synesis.com.au
 *
 *              This source code is placed into the public domain 2003
 *              by Synesis Software Pty Ltd. There are no restrictions
 *              whatsoever to your use of the software.
 *
 *              This source code is provided by Synesis Software Pty Ltd "as is"
 *              and any warranties, whether expressed or implied, including, but
 *              not limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software Pty Ltd be liable for any direct,
 *              indirect, incidental, special, exemplary, or consequential
 *              damages (including, but not limited to, procurement of
 *              substitute goods or services; loss of use, data, or profits; or
 *              business interruption) however caused and on any theory of
 *              liability, whether in contract, strict liability, or tort
 *              (including negligence or otherwise) arising in any way out of
 *              the use of this software, even if advised of the possibility of
 *              such damage.
 *
 *              Neither the name of Synesis Software Pty Ltd nor the names of
 *              any subdivisions, employees or agents of Synesis Software Pty
 *              Ltd, nor the names of any other contributors to this software
 *              may be used to endorse or promote products derived from this
 *              software without specific prior written permission.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/* STLSoft Header Files */
#include <stlsoft.h>

/* PlatformSTL Header Files */
#include <platformstl/platformstl.hpp>
#include <platformstl/file_path_buffer.hpp>
#include <platformstl/filesystem_traits.hpp>

/* Platform-specific Header Files */
#if defined(PLATFORMSTL_OS_IS_UNIX)
# include <unixstl/glob_sequence.h>
#elif defined(PLATFORMSTL_OS_IS_WIN32)
# include <winstl/error_desc.hpp>
# include <winstl/findfile_sequence.hpp>
#else /* ? __SYNSOFT_DVS_OS */
# error Operating system not discriminated
#endif /* __SYNSOFT_DVS_OS */

/* Standard C Header Files */
#include <stdio.h>
#include <string.h>				// for strerror()

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef _SYNSOFT_INTERNAL_BUILD
# include <MLCompId.h>					// The _xxxSynesisSoftware constants
#else /* ? _SYNSOFT_INTERNAL_BUILD */
# define _nameSynesisSoftware				"Synesis Software (Pty) Ltd"
# define _wwwSynesisSoftware_SystemTools	"http://synesis.com.au/"
# define _emailSynesisSoftware_SystemTools	"software@synesis.com.au"
#endif /* _SYNSOFT_INTERNAL_BUILD */

/* /////////////////////////////////////////////////////////////////////////////
 * Version information
 *
 * Normally, Synesis Software components & tools are "versioned" by making use
 * of tool-generated header files, one of which - MBldHdr.h - defines symbolic
 * constants representing major, minor, revision and build numbers.
 *
 * Since this file is to be available in a more public forum, including on the
 * Digital Mars site, these constants are merely defined here, rather than
 * encumber any developers who may wish to compile it with a lot of
 * Synesis-specific gunk.
 */

#ifdef _SYNSOFT_INTERNAL_BUILD
# include "MBldHdr.h"
# include "showver.h"

 const int	_mccVerHi	=	__SYV_MAJOR;
 const int	_mccVerLo	=	__SYV_MINOR;
 const int	_mccVerRev	=	__SYV_REVISION;
 const int	_mccBldNum	=	__SYV_BUILDNUMBER;

#else /* ? _SYNSOFT_INTERNAL_BUILD */
 const int	_mccVerHi	=	1;
 const int	_mccVerLo	=	0;
 const int	_mccVerRev	=	5;
 const int	_mccBldNum	=	6;
#endif /* _SYNSOFT_INTERNAL_BUILD */

#define	COMPANY_NAME	"Synesis Software"
#define	TOOL_NAME		"Empty Directory Removal"
#define	EXE_NAME		"rdempty"

/* /////////////////////////////////////////////////////////////////////////////
 * Function declarations
 */

static void usage(int bExit, char const *reason);
static bool directory_is_empty(char const *dir);
static int empty_directory(char const *dir, int bVerbose, int bRemoveReadOnly);

/* /////////////////////////////////////////////////////////////////////////////
 * Typedefs
 */

typedef char										char_t;
typedef platformstl::filesystem_traits<char_t>		traits_t;
typedef platformstl::basic_file_path_buffer<char_t>	buffer_t;

#if defined(PLATFORMSTL_OS_IS_UNIX)
typedef unixstl::glob_sequence						find_seq_t;
#elif defined(PLATFORMSTL_OS_IS_WIN32)
typedef winstl::basic_findfile_sequence<char_t>		find_seq_t;
#else /* ? __SYNSOFT_DVS_OS */
# error Operating system not discriminated
#endif /* __SYNSOFT_DVS_OS */

/* /////////////////////////////////////////////////////////////////////////////
 * Constants
 */

#define WILDCARDS_ALL		traits_t::pattern_all()

/* ////////////////////////////////////////////////////////////////////////// */

int main(int argc, char *argv[])
{
    int			iRet			=   0;
    int			i;
	char const	*root			=	NULL;
	buffer_t	root_;
	char		*p;
	int			bVerbose		=	0;
	int			bRemoveReadOnly	=	0;

#ifdef _DEBUG
	puts("rdempty: " __STLSOFT_COMPILER_LABEL_STRING);
#endif /* _DEBUG */

	for(i = 1; i < argc; ++i)
	{
		const char	*arg	=	argv[i];

		if(arg[0] == '-')
		{
			if(0 == strcmp("-?", arg))
			{
				usage(1, NULL);
			}
#ifdef _SYNSOFT_INTERNAL_BUILD
			else if(0 == strcmp("--version", arg))
			{
				return show_version();
			}
#endif /* _SYNSOFT_INTERNAL_BUILD */
			else if(0 == strcmp("-v", arg))
			{
				bVerbose = 1;
			}
			else if(0 == strcmp("-r", arg))
			{
				bRemoveReadOnly = 1;
			}
			else
			{
				usage(1, "Invalid argument(s) specified");
			}
		}
		else
		{
			/* other arguments */
			if(NULL == root)
			{
				root = arg;
			}
			else
			{
				usage(1, "Invalid argument(s) specified");
			}
		}
	}

	/* . */
	if(NULL == root)
	{
		root = ".";
	}

	if(!traits_t::get_full_path_name(root, root_.size(), &root_[0], &p))
	{
		printf("Directory \"%s\" does not exist!\n", root);
	}
	else
	{
		root = stlsoft::c_str_ptr(root_);

		if(bVerbose)
		{
			printf("Cleaning directory \"%s\"\n", root);
		}

		int	cRemoved	=	empty_directory(root, bVerbose, bRemoveReadOnly);

		if(bVerbose)
		{
			printf("Removed %d empty directories\n", cRemoved);
		}
	}

    return iRet;
}

/* ////////////////////////////////////////////////////////////////////////// */

bool directory_is_empty(char const *dir)
{
	return find_seq_t(dir, WILDCARDS_ALL, find_seq_t::files | find_seq_t:: directories).empty();
}

int empty_directory(char const *dir, int bVerbose, int bRemoveReadOnly)
{
	int	cRemoved	=	0;

	try
	{
		find_seq_t					dirs(dir, WILDCARDS_ALL, find_seq_t::directories);
		find_seq_t::const_iterator	begin = dirs.begin();
		find_seq_t::const_iterator	end		= dirs.end();

		for(; begin != end; ++begin)
		{
			cRemoved += empty_directory(stlsoft::c_str_ptr(*begin), bVerbose, bRemoveReadOnly);
		}

		if(directory_is_empty(dir))
		{
#if defined(PLATFORMSTL_OS_IS_WIN32)
			DWORD	attr	=	::GetFileAttributesA(dir);

			stlsoft_assert(0 != (FILE_ATTRIBUTE_DIRECTORY & attr));

			if( 0 != (FILE_ATTRIBUTE_READONLY & attr) &&
				bRemoveReadOnly)
			{
				if(!::SetFileAttributes(dir, attr & ~(FILE_ATTRIBUTE_READONLY)))
				{
					goto err;
				}
			}
#endif /* WIN32 */

			if(!traits_t::remove_directory(dir))
			{
				++cRemoved;

				if(bVerbose)
				{
					printf("Removed \"%s\"\n", dir);
				}
			}
			else
			{
#if defined(PLATFORMSTL_OS_IS_WIN32)
err:

				printf("Could not remove \"%s\": %s\n", dir, (char const*)::winstl::error_desc(::GetLastError()));
#else /* ? WIN32 */
				printf("Could not remove \"%s\": %s\n", dir, strerror(errno));
#endif /* WIN32 */
			}
		}
	}
	catch(std::exception &x)
	{
		fprintf(stderr, "Error occured when processing directory: %s; %s\n", dir, x.what());
	}

	return cRemoved;
}

void usage(int bExit, char const *reason)
{
	FILE	*stm	=	(NULL == reason) ? stdout : stderr;

    fprintf(stm, "\n");
	fprintf(stm, "  %s %s Tool, v%d.%d.%d.%04d\n", COMPANY_NAME, TOOL_NAME, _mccVerHi, _mccVerLo, _mccVerRev, _mccBldNum);
    fprintf(stm, "\n");

	if(NULL != reason)
	{
		fprintf(stm, "  Error: %s\n", reason);
		fprintf(stm, "\n");
	}

    fprintf(stm, "  USAGE 1: %s [-v] [<dir>]\n", EXE_NAME);
    fprintf(stm, "    where:\n\n");
    fprintf(stm, "    <dir>          - root directory in search (defaults to current directory)\n");
    fprintf(stm, "    -r             - removes empty directories marked read-only\n");
    fprintf(stm, "    -v             - prints out names of deleted directories, and total deleted\n");
    fprintf(stm, "\n");
    fprintf(stm, "    Removes all empty sub-directories of the given directory\n");
    fprintf(stm, "\n");
    fprintf(stm, "  USAGE 2: %s -?\n", EXE_NAME);
    fprintf(stm, "\n");
    fprintf(stm, "    Displays this help\n");
    fprintf(stm, "\n");
	fprintf(stm, "  Contact %s\n", _nameSynesisSoftware);
	fprintf(stm, "    at \"%s\",\n", _wwwSynesisSoftware_SystemTools);
	fprintf(stm, "    or, via email, at \"%s\"\n", _emailSynesisSoftware_SystemTools);
    fprintf(stm, "\n");

    if(bExit)
    {
        exit(EXIT_FAILURE);
    }
}

/* ////////////////////////////////////////////////////////////////////////// */
